<?php
/**
 * @file
 * Provides functionality associated with Secure Hosting basic offsite payment
 * services, as described in 'Integration with SecureHosting.doc'.
 * 
 * @see http://www.secure-server-hosting.com/secutran/documentation/
 */

/**
 * The path to use as the action on live transactions.
 */
define('COMMERCE_SECUREHOSTING_OB_LIVE_SERVER', 'https://www.secure-server-hosting.com/secutran/secuitems.php');

/**
 * The path to use as the action on test transactions.
 */
define('COMMERCE_SECUREHOSTING_OB_TEST_SERVER', 'https://test.secure-server-hosting.com/secutran/secuitems.php');

/**
 * The path to use to get the secustring field.
 */
define('COMMERCE_SECUREHOSTING_OB_SECUSTRING_SERVER', 'https://www.secure-server-hosting.com/secutran/create_secustring.php');

/**
 * The HTTP referer sent by the Secure Hosting server when calling our callback.
 */
define('COMMERCE_SECUREHOSTING_CALLBACK_REFERER', 'https://www.secure-server-hosting.com/secutran/ProcessCallbacks.php');

/**
 * The HTTP referer we send when contacting Secure Hosting to encrypt fields as
 * a Drupal path.
 */
define('COMMERCE_SECUREHOSTING_ENCRYPTION_REFERER', 'commerce_securehosting/referer');

/**
 * Returns the default settings for the Authorize.Net AIM payment method.
 */
function commerce_securehosting_ob_default_settings() {
  return array(
    'sh_ref' => 'sh',
    'check_code' => '',
    'server' => 'test',
    'shared_secret' => '',
    'secuitems_encrypt_phrase' => '',
  );
}

/**
 * Payment method callback: settings form.
 */
function commerce_securehosting_ob_settings_form($settings = NULL) {
  // Merge default settings into the stored settings array.
  $settings = (array) $settings + commerce_securehosting_ob_default_settings();

  $form = array();

  $form['sh_ref'] = array(
    '#type' => 'textfield',
    '#title' => t('Secure Hosting client reference number'),
    '#description' => t("This is your account's reference number. It should start with <em>sh</em>."),
    '#default_value' => $settings['sh_ref'],
    '#required' => TRUE,
  );
  $form['check_code'] = array(
    '#type' => 'textfield',
    '#title' => t('Check code'),
    '#description' => t("Your check code was issued to you when you set up the account."),
    '#default_value' => $settings['check_code'],
    '#required' => TRUE,
  );
  $form['shared_secret'] = array(
    '#type' => 'textfield',
    '#title' => t('Shared secret'),
    '#description' => t("In order to verify Secure Hosting's payment notification does indeed come from Secure Hosting, it's necessary to enter the same shared secret here and on the Secure Hosting site (Settings >> Advanced Settings >> Shared Secret)"),
    '#default_value' => $settings['shared_secret'],
    '#required' => TRUE,
  );
  $form['secuitems_encrypt_phrase'] = array(
    '#type' => 'textfield',
    '#title' => t('Advanced Secuitems encryption phrase'),
    '#description' => t("In order to prevent tampering with the details passed to Secure Hosting it's necessary to enter the same 6-9 character phrase here and on the Secure Hosting site (Settings >> Advanced Settings >> A phrase to be used to further encrypt your data sent through secuitems)."),
    '#default_value' => $settings['secuitems_encrypt_phrase'],
    '#required' => TRUE,
  );
  $form['server'] = array(
    '#type' => 'radios',
    '#title' => t('Secure Hosting server'),
    '#options' => array(
      'test' => ('Sandbox - use for testing'),
      'live' => ('Live - use for processing real transactions'),
    ),
    '#description' => t("<strong>Please note that the form on template.html (which you have to upload to your Secure Hosting account) also needs to use the testing server.</strong> If you remove the domain name from the action attribute it should work fine for live and test use."),
    '#default_value' => $settings['server'],
  );
  $form['callback_url'] = array(
    '#type' => 'item',
    '#title' => t('Callback URL'),
    '#markup' => url(COMMERCE_SECUREHOSTING_CALLBACK_URL, array('absolute' => TRUE)),
    '#description' => t("This should be entered on the Advanced Settings screen (Settings >> Advanced Settings >> Callback URL)"),
  );
  $form['cart_referer'] = array(
    '#type' => 'item',
    '#title' => t("Shopping cart referer"),
    '#markup' => _commerce_securehost_ob_get_encryption_referer(),
    '#description' => t("This should be entered on the Advanced Settings screen (Settings >> Advanced Settings >> The full URL referrer of your shopping cart)"),
  );

  return $form;
}

/**
 * Payment method callback: redirect form. Creates the form for passing the
 * order to Secure Hosting.
 * 
 * @see _commerce_securehost_ob_get_order_string()
 * @see _commerce_securehost_ob_get_line_items()
 * @see _commerce_securehost_ob_get_secustring()
 * @see _commerce_securehost_ob_get_addresses()
 */
function commerce_securehosting_ob_redirect_form($form, &$form_state, $order, $payment_method) {

  // Return an error if the payment method's settings haven't been configured.
  $settings = $payment_method['settings'];
  if (!_commerce_securehosting_ob_check_settings($settings)) {
    drupal_set_message(t('The Secure Hosting payment method is not configured correctly for use.'), 'error');
    return array();
  }

  $wrapper = entity_metadata_wrapper('commerce_order', $order);
  
  $currency_code = $wrapper->commerce_order_total->currency_code->value();
  $order_string = _commerce_securehost_ob_get_order_string($wrapper, $currency_code);
  
  // Get the fields we want to be secure.
  $fields = array(
    'shreference' => $settings['sh_ref'],
    'checkcode' => $settings['check_code'],
    'secuitems' => $order_string,
    'transactioncurrency' => $currency_code,
    'orderid' => $order->order_id,
    'cancel_return' => url('checkout/' . $order->order_id . '/payment/back/' . $order->data['payment_redirect_key'], array('absolute' => TRUE)),
    'return' => url('checkout/' . $order->order_id . '/payment/return/' . $order->data['payment_redirect_key'], array('absolute' => TRUE)),
  );
  $fields += _commerce_securehost_ob_get_line_items($wrapper, $currency_code);

  // Get the hash for the secure fields.
  $secustring = _commerce_securehost_ob_get_secustring($fields, $payment_method);
  
  // Add secure field hash and fields that don't need to be secure.
  $fields += array(
    'secuString' => $secustring,
    'filename' => $settings['sh_ref'] . '/template.html',
    'store_name' => variable_get('site_name', 'Drupal'),
    'email' => $order->mail,
  );
  $fields += _commerce_securehost_ob_get_addresses($wrapper);
  
  // Set the appropriate action
  switch ($settings['server']) {
    case 'live':
      $form['#action'] = COMMERCE_SECUREHOSTING_OB_LIVE_SERVER;
      break;
    case 'test':
      $form['#action'] = COMMERCE_SECUREHOSTING_OB_TEST_SERVER;
      break; 
  }

  foreach ($fields as $name => $value) {
    $form[$name] = array(
      '#type' => 'hidden',
      '#value' => $value,
    );
  }
  
  $form['submit'] = array(
    '#type' => 'submit',
    '#value' => t('Proceed to pay off-site'),
  );
  
  return $form;
}

/**
 * Callback handler for Secure Hosting's payment notification.
 */
function commerce_securehosting_ob_callback() {

  // Check referer.
  if (!isset($_SERVER['HTTP_REFERER']) || ($_SERVER['HTTP_REFERER'] != COMMERCE_SECUREHOSTING_CALLBACK_REFERER)) {
    drupal_exit();
  }
  
  // Check required GET parameters.
  $get_fields = array(
    'orderid',
    'verify',
    'transactionnumber'
  );
  $missing_fields = array();
  foreach ($get_fields as $field) {
    if (!isset($_GET[$field]) || !$_GET[$field]) {
      $missing_fields[] = $field;
    }
  }
  if ($missing_fields) {
    watchdog('commerce_securehosting', 'Callback not processed: Request was missing required fields: %fields.', array('%fields' => implode(', ', $missing_fields)), WATCHDOG_ALERT);
    drupal_exit();
  }

  // Load the order.
  $order = commerce_order_load($_GET['orderid']);
  if (!$order) {
    watchdog('commerce_securehosting', "Callback not processed: Request had invalid order ID.", array(), WATCHDOG_ALERT);
    drupal_exit();
  }
  
  // Load the payment method.
  $payment_method = commerce_payment_method_instance_load($order->data['payment_method']);
  if (!$payment_method) {
    watchdog('commerce_securehosting', 'Callback not processed: Unable to load payment method %method.', array('%method' => $instance_id), WATCHDOG_ALERT);
    drupal_exit();
  }
  
  // Check the payment method's settings seem valid.
  $settings = $payment_method['settings'];
  if (!_commerce_securehosting_ob_check_settings($settings)) {
    watchdog('commerce_securehosting', 'Callback not processed: Invalid payment method settings.', array(), WATCHDOG_ALERT);
    drupal_exit();
  }
  
  // Check that the notification's valid via the shared secret.
  $salt = $settings['shared_secret'];
  $tx_num = $_GET['transactionnumber'];
  $verify = $_GET['verify'];
  
  $check = sha1($salt . $tx_num . $salt);
  if ($check != $verify) {
    watchdog('commerce_securehosting', "Callback not processed: Request has invalid verification number. (Maybe the shared secrete is incorrectly configured on this site or Secure Hosting.)", array(), WATCHDOG_ALERT);
    drupal_exit();
  }
  
  // At this point we've checked that the verification number is the transaction
  // number encrypted with SHA1 and salted with the shared secret entered for
  // this order's payment method.
  
  // Do I need to check the order status here? Seems sensible, but I don't
  // notice commerce_paypal_wps doing it.
  
  // Create a new transaction for this notification.
  $wrapper = entity_metadata_wrapper('commerce_order', $order);
  
  $transaction = commerce_payment_transaction_new('securehosting_ob', $order->order_id);
  $transaction->instance_id = $payment_method['instance_id'];
  
  $transaction->remote_id = $_GET['transactionnumber'];

  // It's important to use the values for amount and currency that were sent to
  // Secure Hosting (which can be trusted because they're included as encrypted
  // fields in SH's config) just in case the order has changed between the time
  // the initial data was sent to SH and now.
  $total = $_GET['transactionamount'];
  $currency_code = $_GET['transactioncurrency'];
  $transaction->amount = commerce_currency_decimal_to_amount($total, $currency_code);
  $transaction->currency_code = $currency_code;

  $payload_fields = array('sh_reason', 'transactiontime', 'upgauthcode');
  foreach ($payload_fields as $field) {
    if (isset($_GET[$field])) {
      $transaction->payload[$field] = $_GET[$field];
    }
  }
  
  // Check if the transaction was a success or not and respond appropriately.
  if ($tx_num == -1) {
    $transaction->status = COMMERCE_PAYMENT_STATUS_FAILURE;
    $transaction->remote_status = COMMERCE_PAYMENT_STATUS_FAILURE;
    $transaction->message = t("Payment failure: @message", array('@message' => $_GET['failurereason']));
    commerce_payment_transaction_save($transaction);
    // Don't use commerce_payment_redirect_pane_previous_page() here because
    // that will cause problems if the customer stays on Secure Hosting and
    // tries to resubmit their details.
  }
  else {
    $transaction->status = COMMERCE_PAYMENT_STATUS_SUCCESS;
    $transaction->remote_status = COMMERCE_PAYMENT_STATUS_SUCCESS;
    $transaction->message = t("The payment has been completed.");
    commerce_payment_transaction_save($transaction);
    commerce_payment_redirect_pane_next_page($order);
  }
    
  watchdog('commerce_securehosting', 'Processed callback for order %order_id.', array('%order_id' => $order->order_id), WATCHDOG_NOTICE);
  
  // The only output from the callback should be 'success'.
  echo 'success';
  drupal_exit();
}

/**
 * Returns the secuString value for the given fields and payment method.
 *
 * @return
 *   The hash to be used as a value in the hidden input secuString if the call
 *   succeeded; FALSE otherwise
 */
function _commerce_securehost_ob_get_secustring($fields, $payment_method) {
  
  $settings = $payment_method['settings'];
  
  $fields += array(
    'secuphrase' => $settings['secuitems_encrypt_phrase'],
  );
  
  $http_referer = _commerce_securehost_ob_get_encryption_referer();
    
  $ch = curl_init();
  curl_setopt ($ch, CURLOPT_URL, COMMERCE_SECUREHOSTING_OB_SECUSTRING_SERVER);
  curl_setopt ($ch, CURLOPT_POST, 1);
  curl_setopt ($ch, CURLOPT_POSTFIELDS, $fields);
  curl_setopt ($ch, CURLOPT_HEADER, 0);
  curl_setopt ($ch, CURLOPT_REFERER, $http_referer); 
  curl_setopt ($ch, CURLOPT_RETURNTRANSFER,1);
  curl_setopt ($ch, CURLOPT_SSL_VERIFYPEER, 0);
  curl_setopt ($ch, CURLOPT_TIMEOUT, 10);
  $secustring = trim(curl_exec($ch));
  curl_close ($ch);

  if ($secustring) {
    // Secure Hosting return the entire hidden input tag - we're only interested
    // in the value.
    $matches = array();
    preg_match('/value="([0-9a-f]+)"/i', $secustring, $matches);
    if (isset($matches[1])) {
      return $matches[1];
    }
  }
  
  return FALSE;
}

/**
 * To use advanced secuitems, you need to login to Secure Hosting and specify
 * the HTTP referer that will be used by the calling script. This function
 * returns the appropriate HTTP referer.
 */
function _commerce_securehost_ob_get_encryption_referer() {
  return url(COMMERCE_SECUREHOSTING_ENCRYPTION_REFERER, array('absolute' => TRUE));
}

/**
 * Returns an order string for use as the secuitems value.
 * 
 * From 'Integration with SecureHosting.doc':
 * <input type="hidden" name="secuitems" value="[itemcode|itemskew|itemdesc|itempric|itemquan|itemtota][itemcode2|itemskew2|itemdesc2|itempric2|itemquan2|itemtota2]">
 * where the values will be inserted :-
 *   itemcode = product code
 *   itemskew = SKU code
 *   itemdesc = product description
 *   itempric = product price
 *   itemquan = quantity
 *   itemtota = line total
 * 
 * As the only optional field is itemskew, and there's only one product code
 * field on normal products (SKU), the SKU will be used as itemcode, and
 * itemskew will be omitted.
 * 
 * @param $order_wrapper
 *   Entity metadata wrapper for the order from which to generate the order
 *   string.
 * 
 * @return
 *   The order string
 * 
 * @see commerce_securehosting_ob_redirect_form()
 */
function _commerce_securehost_ob_get_order_string($order_wrapper, $currency_code) {
  
  $order_string = '';
  foreach ($order_wrapper->commerce_line_items as $delta => $line_item_wrapper) {
    if (in_array($line_item_wrapper->type->value(), commerce_product_line_item_types())) {
      
      // [itemcode|itemskew|itemdesc|itempric|itemquan|itemtota]
      // [SKU||title|price|quantity|total]
      
      $commerce_product = $line_item_wrapper->commerce_product;
      
      $order_string .= '[';

      $order_string .= $commerce_product->sku->value() . '|';
      $order_string .= '|'; // Skip 'itemskew'
      $order_string .= $commerce_product->title->value() . '|';
      $price = $line_item_wrapper->commerce_unit_price->value();
      $order_string .= _commerce_securehosting_ob_currency_amount_to_decimal_string($price, $currency_code) . '|';
      $order_string .= round($line_item_wrapper->quantity->value(), 2) . '|';
      $total = $line_item_wrapper->commerce_total->value();
      $order_string .= _commerce_securehosting_ob_currency_amount_to_decimal_string($total, $currency_code);
      
      $order_string .= ']';
    }
  }
  
  return $order_string;
}

/**
 * Returns an array of the order's line items (not products).
 * 
 * @param $order_wrapper
 *   Entity metadata wrapper for order
 * 
 * @return
 *   Array of line item totals keyed by the appropriate Secure Hosting field
 *   name.
 */
function _commerce_securehost_ob_get_line_items($order_wrapper, $currency_code) {
  
  $line_items = array();
  $order_total = $order_wrapper->commerce_order_total->value();

  // Subtotal
  $base_price = commerce_price_component_load($order_total, 'base_price');
  // I'm presuming there's no need to total 'base_price' components - surely
  // there should be only one?
  $base_price = $base_price[0]['price'];
  $subtotal = _commerce_securehosting_ob_currency_amount_to_decimal_string($base_price, $currency_code);
  $line_items['products_price'] = $subtotal;
  
  // Shipping
  foreach ($order_wrapper->commerce_line_items as $delta => $line_item_wrapper) {
    switch ($line_item_wrapper->type->value()) {
      case 'shipping':
        $amount = $line_item_wrapper->commerce_total->value();
        $shipping = _commerce_securehosting_ob_currency_amount_to_decimal_string($amount, $currency_code);
        $line_items['shippingcharge'] = $shipping;
        break;
    }
  }
  
  // Tax
  if (function_exists('commerce_tax_total_amount')) {
    // This returns an amount already converted to $currency_code.
    $tax = commerce_tax_total_amount($order_total['data']['components'], FALSE, $currency_code);
    $line_items['transactiontax'] = _commerce_securehosting_ob_currency_amount_to_decimal_string($tax, $currency_code);
  }
  
  // Total
  $total = _commerce_securehosting_ob_currency_amount_to_decimal_string($order_total, $currency_code);
  $line_items['transactionamount'] = $total;
  
  return $line_items;
}

/**
 * Returns an array of address items from an order.
 * 
 * @param $order_wrapper
 *   Entity metadata wrapper for order
 *
 * @return
 *   Array of address values keyed by the Secure Hosting field name.
 */
function _commerce_securehost_ob_get_addresses($order_wrapper) {
  $fields = array();

  if (isset($order_wrapper->commerce_customer_billing->commerce_customer_address)) {
    $billing = $order_wrapper->commerce_customer_billing->commerce_customer_address->value();
    // These fields come from 'Integration with SecureHosting.doc'.
    $billing_map = array(
      'cardholdersname' => 'name_line',
      'cardHolderAddr1' => 'thoroughfare',
      'cardHolderCity' => 'locality',
      'cardHolderState' => 'administrative_area',
      'cardHolderPostcode' => 'postal_code',
    );
    foreach ($billing_map as $sh_name => $address_name) {
      if (isset($billing[$address_name])) {
        $fields[$sh_name] = $billing[$address_name];
      }
    }
  }
  
  if (isset($order_wrapper->commerce_customer_shipping->commerce_customer_address)) {
    
    $shipping = $order_wrapper->commerce_customer_shipping->commerce_customer_address->value();
    // These fields come from Secure Hosting's example template.html.
    $shipping_map = array(
      'deliveryname' => 'name_line',
      'deliveryAddr1' => 'thoroughfare',
      'deliveryCity' => 'locality',
      'deliveryState' => 'administrative_area',
      'deliveryPostcode' => 'postal_code',
    );
    foreach ($shipping_map as $sh_name => $address_name) {
      if (isset($shipping[$address_name])) {
        $fields[$sh_name] = $shipping[$address_name];
      }
    }
  }
  return $fields;
}

/**
 * Does a quick offline check of settings to ensure they appear to be in order.
 * 
 * @param $settings
 *   Settings array
 *
 * @return
 *    TRUE if the settings appear good; FALSE otherwise
 */
function _commerce_securehosting_ob_check_settings($settings) {
  if (($settings['sh_ref'] == 'sh') || empty($settings['sh_ref'])) {
    return FALSE;
  }
  if (empty($settings['check_code'])) {
    return FALSE;
  }
  if (!in_array($settings['server'], array('live', 'test'))) {
    return FALSE;
  }
  if (empty($settings['shared_secret'])) {
    return FALSE;
  }
  if (empty($settings['secuitems_encrypt_phrase'])) {
    return FALSE;
  }
  return TRUE;
}

/**
 * Converts a price amount to a decimal string appropriate for use in the value
 * attribute of the secuitems hidden input based on the currency.
 * 
 * 25/5/2012 Darren Potter of UPG confirmed that when passing prices to Secure
 * Hosting there should be a decimal point, no thousand separator, and a maximum
 * of two decimal places after the point.
 * 
 * @param $price
 *   The price to convert to a decimal string: can be numerical, or a price
 *   array (in which case it will be converted to $currency_code)
 * @param $currency_code
 *   If $price is a price array then $currency_code is the target currency code;
 *   otherwise $currency_code indicates the currency $price is in.
 * 
 * @return
 *   The decimal amount formatted with two numbers after the decimal point as
 *   a string.
 */
function _commerce_securehosting_ob_currency_amount_to_decimal_string($price, $currency_code) {
  if (is_array($price)) {
    $price = commerce_currency_convert($price['amount'], $price['currency_code'], $currency_code);
  }
  $decimal_price = commerce_currency_amount_to_decimal($price, $currency_code);
  return number_format($decimal_price, 2, '.', '');
}
